/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "NetworkStorageManager.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "FileSystemStorageError.h"
#include "HandleMessage.h"
#include "NetworkStorageManagerMessages.h"
#include "SharedFileHandle.h"
#include "WebCoreArgumentCoders.h"
#include <WebCore/ClientOrigin.h>
#include <WebCore/FileSystemHandleIdentifier.h>
#include <WebCore/FileSystemSyncAccessHandleIdentifier.h>
#include <optional>
#include <utility>
#include <wtf/Expected.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace NetworkStorageManager {

void Persisted::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    std::optional<bool> persisted;
    decoder >> persisted;
    if (!persisted) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*persisted));
}

void Persisted::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<bool>::create());
}

void Persisted::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, bool persisted)
{
    encoder.get() << persisted;
    connection.sendSyncReply(WTFMove(encoder));
}

void Persist::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    std::optional<bool> persisted;
    decoder >> persisted;
    if (!persisted) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*persisted));
}

void Persist::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<bool>::create());
}

void Persist::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, bool persisted)
{
    encoder.get() << persisted;
    connection.sendSyncReply(WTFMove(encoder));
}

void FileSystemGetDirectory::callReply(IPC::Decoder& decoder, CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    std::optional<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void FileSystemGetDirectory::cancelReply(CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>>::create());
}

void FileSystemGetDirectory::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>& result)
{
    encoder.get() << result;
    connection.sendSyncReply(WTFMove(encoder));
}

void IsSameEntry::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    std::optional<bool> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void IsSameEntry::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<bool>::create());
}

void IsSameEntry::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, bool result)
{
    encoder.get() << result;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetFileHandle::callReply(IPC::Decoder& decoder, CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    std::optional<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void GetFileHandle::cancelReply(CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>>::create());
}

void GetFileHandle::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>& result)
{
    encoder.get() << result;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetDirectoryHandle::callReply(IPC::Decoder& decoder, CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    std::optional<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void GetDirectoryHandle::cancelReply(CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>>::create());
}

void GetDirectoryHandle::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>& result)
{
    encoder.get() << result;
    connection.sendSyncReply(WTFMove(encoder));
}

void RemoveEntry::callReply(IPC::Decoder& decoder, CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    std::optional<std::optional<WebKit::FileSystemStorageError>> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void RemoveEntry::cancelReply(CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<std::optional<WebKit::FileSystemStorageError>>::create());
}

void RemoveEntry::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const std::optional<WebKit::FileSystemStorageError>& result)
{
    encoder.get() << result;
    connection.sendSyncReply(WTFMove(encoder));
}

void Resolve::callReply(IPC::Decoder& decoder, CompletionHandler<void(Expected<Vector<String>, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    std::optional<Expected<Vector<String>, WebKit::FileSystemStorageError>> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void Resolve::cancelReply(CompletionHandler<void(Expected<Vector<String>, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Expected<Vector<String>, WebKit::FileSystemStorageError>>::create());
}

void Resolve::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const Expected<Vector<String>, WebKit::FileSystemStorageError>& result)
{
    encoder.get() << result;
    connection.sendSyncReply(WTFMove(encoder));
}

void Move::callReply(IPC::Decoder& decoder, CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    std::optional<std::optional<WebKit::FileSystemStorageError>> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void Move::cancelReply(CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<std::optional<WebKit::FileSystemStorageError>>::create());
}

void Move::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const std::optional<WebKit::FileSystemStorageError>& result)
{
    encoder.get() << result;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetFile::callReply(IPC::Decoder& decoder, CompletionHandler<void(Expected<String, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    std::optional<Expected<String, WebKit::FileSystemStorageError>> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void GetFile::cancelReply(CompletionHandler<void(Expected<String, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Expected<String, WebKit::FileSystemStorageError>>::create());
}

void GetFile::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const Expected<String, WebKit::FileSystemStorageError>& result)
{
    encoder.get() << result;
    connection.sendSyncReply(WTFMove(encoder));
}

void CreateSyncAccessHandle::callReply(IPC::Decoder& decoder, CompletionHandler<void(Expected<std::pair<WebCore::FileSystemSyncAccessHandleIdentifier, IPC::SharedFileHandle>, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    std::optional<Expected<std::pair<WebCore::FileSystemSyncAccessHandleIdentifier, IPC::SharedFileHandle>, WebKit::FileSystemStorageError>> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void CreateSyncAccessHandle::cancelReply(CompletionHandler<void(Expected<std::pair<WebCore::FileSystemSyncAccessHandleIdentifier, IPC::SharedFileHandle>, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Expected<std::pair<WebCore::FileSystemSyncAccessHandleIdentifier, IPC::SharedFileHandle>, WebKit::FileSystemStorageError>>::create());
}

void CreateSyncAccessHandle::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const Expected<std::pair<WebCore::FileSystemSyncAccessHandleIdentifier, IPC::SharedFileHandle>, WebKit::FileSystemStorageError>& result)
{
    encoder.get() << result;
    connection.sendSyncReply(WTFMove(encoder));
}

void CloseAccessHandle::callReply(IPC::Decoder& decoder, CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    std::optional<std::optional<WebKit::FileSystemStorageError>> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void CloseAccessHandle::cancelReply(CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<std::optional<WebKit::FileSystemStorageError>>::create());
}

void CloseAccessHandle::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const std::optional<WebKit::FileSystemStorageError>& result)
{
    encoder.get() << result;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetHandleNames::callReply(IPC::Decoder& decoder, CompletionHandler<void(Expected<Vector<String>, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    std::optional<Expected<Vector<String>, WebKit::FileSystemStorageError>> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void GetHandleNames::cancelReply(CompletionHandler<void(Expected<Vector<String>, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Expected<Vector<String>, WebKit::FileSystemStorageError>>::create());
}

void GetHandleNames::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const Expected<Vector<String>, WebKit::FileSystemStorageError>& result)
{
    encoder.get() << result;
    connection.sendSyncReply(WTFMove(encoder));
}

void GetHandle::callReply(IPC::Decoder& decoder, CompletionHandler<void(Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    std::optional<Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void GetHandle::cancelReply(CompletionHandler<void(Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>>::create());
}

void GetHandle::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>& result)
{
    encoder.get() << result;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace NetworkStorageManager

} // namespace Messages

namespace WebKit {

void NetworkStorageManager::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    Ref protectedThis { *this };
    if (decoder.messageName() == Messages::NetworkStorageManager::Persisted::name())
        return IPC::handleMessageAsync<Messages::NetworkStorageManager::Persisted>(connection, decoder, this, &NetworkStorageManager::persisted);
    if (decoder.messageName() == Messages::NetworkStorageManager::Persist::name())
        return IPC::handleMessageAsync<Messages::NetworkStorageManager::Persist>(connection, decoder, this, &NetworkStorageManager::persist);
    if (decoder.messageName() == Messages::NetworkStorageManager::FileSystemGetDirectory::name())
        return IPC::handleMessageAsyncWantsConnection<Messages::NetworkStorageManager::FileSystemGetDirectory>(connection, decoder, this, &NetworkStorageManager::fileSystemGetDirectory);
    if (decoder.messageName() == Messages::NetworkStorageManager::CloseHandle::name())
        return IPC::handleMessage<Messages::NetworkStorageManager::CloseHandle>(connection, decoder, this, &NetworkStorageManager::closeHandle);
    if (decoder.messageName() == Messages::NetworkStorageManager::IsSameEntry::name())
        return IPC::handleMessageAsync<Messages::NetworkStorageManager::IsSameEntry>(connection, decoder, this, &NetworkStorageManager::isSameEntry);
    if (decoder.messageName() == Messages::NetworkStorageManager::GetFileHandle::name())
        return IPC::handleMessageAsyncWantsConnection<Messages::NetworkStorageManager::GetFileHandle>(connection, decoder, this, &NetworkStorageManager::getFileHandle);
    if (decoder.messageName() == Messages::NetworkStorageManager::GetDirectoryHandle::name())
        return IPC::handleMessageAsyncWantsConnection<Messages::NetworkStorageManager::GetDirectoryHandle>(connection, decoder, this, &NetworkStorageManager::getDirectoryHandle);
    if (decoder.messageName() == Messages::NetworkStorageManager::RemoveEntry::name())
        return IPC::handleMessageAsync<Messages::NetworkStorageManager::RemoveEntry>(connection, decoder, this, &NetworkStorageManager::removeEntry);
    if (decoder.messageName() == Messages::NetworkStorageManager::Resolve::name())
        return IPC::handleMessageAsync<Messages::NetworkStorageManager::Resolve>(connection, decoder, this, &NetworkStorageManager::resolve);
    if (decoder.messageName() == Messages::NetworkStorageManager::Move::name())
        return IPC::handleMessageAsync<Messages::NetworkStorageManager::Move>(connection, decoder, this, &NetworkStorageManager::move);
    if (decoder.messageName() == Messages::NetworkStorageManager::GetFile::name())
        return IPC::handleMessageAsync<Messages::NetworkStorageManager::GetFile>(connection, decoder, this, &NetworkStorageManager::getFile);
    if (decoder.messageName() == Messages::NetworkStorageManager::CreateSyncAccessHandle::name())
        return IPC::handleMessageAsync<Messages::NetworkStorageManager::CreateSyncAccessHandle>(connection, decoder, this, &NetworkStorageManager::createSyncAccessHandle);
    if (decoder.messageName() == Messages::NetworkStorageManager::CloseAccessHandle::name())
        return IPC::handleMessageAsync<Messages::NetworkStorageManager::CloseAccessHandle>(connection, decoder, this, &NetworkStorageManager::closeAccessHandle);
    if (decoder.messageName() == Messages::NetworkStorageManager::GetHandleNames::name())
        return IPC::handleMessageAsync<Messages::NetworkStorageManager::GetHandleNames>(connection, decoder, this, &NetworkStorageManager::getHandleNames);
    if (decoder.messageName() == Messages::NetworkStorageManager::GetHandle::name())
        return IPC::handleMessageAsyncWantsConnection<Messages::NetworkStorageManager::GetHandle>(connection, decoder, this, &NetworkStorageManager::getHandle);
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
#if ENABLE(IPC_TESTING_API)
    if (connection.ignoreInvalidMessageForTesting())
        return;
#endif // ENABLE(IPC_TESTING_API)
    ASSERT_NOT_REACHED_WITH_MESSAGE("Unhandled message %s to %" PRIu64, description(decoder.messageName()), decoder.destinationID());
}

} // namespace WebKit
