/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "Connection.h"
#include "FileSystemStorageError.h"
#include "MessageNames.h"
#include "NetworkStorageManagerMessagesReplies.h"
#include "SharedFileHandle.h"
#include <WebCore/FileSystemHandleIdentifier.h>
#include <WebCore/FileSystemSyncAccessHandleIdentifier.h>
#include <optional>
#include <utility>
#include <wtf/Expected.h>
#include <wtf/Forward.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace WebCore {
struct ClientOrigin;
}

namespace Messages {
namespace NetworkStorageManager {

static inline IPC::ReceiverName messageReceiverName()
{
    return IPC::ReceiverName::NetworkStorageManager;
}

class Persisted {
public:
    using Arguments = std::tuple<const WebCore::ClientOrigin&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_Persisted; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_PersistedReply; }
    using AsyncReply = PersistedAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool persisted);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit Persisted(const WebCore::ClientOrigin& origin)
        : m_arguments(origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Persist {
public:
    using Arguments = std::tuple<const WebCore::ClientOrigin&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_Persist; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_PersistReply; }
    using AsyncReply = PersistAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool persisted);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit Persist(const WebCore::ClientOrigin& origin)
        : m_arguments(origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class FileSystemGetDirectory {
public:
    using Arguments = std::tuple<const WebCore::ClientOrigin&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_FileSystemGetDirectory; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_FileSystemGetDirectoryReply; }
    using AsyncReply = FileSystemGetDirectoryAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>& result);
    using Reply = std::tuple<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>>;
    explicit FileSystemGetDirectory(const WebCore::ClientOrigin& origin)
        : m_arguments(origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CloseHandle {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_CloseHandle; }
    static constexpr bool isSync = false;

    explicit CloseHandle(const WebCore::FileSystemHandleIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class IsSameEntry {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const WebCore::FileSystemHandleIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_IsSameEntry; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_IsSameEntryReply; }
    using AsyncReply = IsSameEntryAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool result);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    IsSameEntry(const WebCore::FileSystemHandleIdentifier& identifier, const WebCore::FileSystemHandleIdentifier& targetIdentifier)
        : m_arguments(identifier, targetIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetFileHandle {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const String&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_GetFileHandle; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_GetFileHandleReply; }
    using AsyncReply = GetFileHandleAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>& result);
    using Reply = std::tuple<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>>;
    GetFileHandle(const WebCore::FileSystemHandleIdentifier& identifier, const String& name, bool createIfNecessary)
        : m_arguments(identifier, name, createIfNecessary)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetDirectoryHandle {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const String&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_GetDirectoryHandle; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_GetDirectoryHandleReply; }
    using AsyncReply = GetDirectoryHandleAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>& result);
    using Reply = std::tuple<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>>;
    GetDirectoryHandle(const WebCore::FileSystemHandleIdentifier& identifier, const String& name, bool createIfNecessary)
        : m_arguments(identifier, name, createIfNecessary)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RemoveEntry {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const String&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_RemoveEntry; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_RemoveEntryReply; }
    using AsyncReply = RemoveEntryAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const std::optional<WebKit::FileSystemStorageError>& result);
    using Reply = std::tuple<std::optional<WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<std::optional<WebKit::FileSystemStorageError>>;
    RemoveEntry(const WebCore::FileSystemHandleIdentifier& identifier, const String& name, bool deleteRecursively)
        : m_arguments(identifier, name, deleteRecursively)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Resolve {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const WebCore::FileSystemHandleIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_Resolve; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<Vector<String>, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<Vector<String>, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_ResolveReply; }
    using AsyncReply = ResolveAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const Expected<Vector<String>, WebKit::FileSystemStorageError>& result);
    using Reply = std::tuple<Expected<Vector<String>, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<Vector<String>, WebKit::FileSystemStorageError>>;
    Resolve(const WebCore::FileSystemHandleIdentifier& identifier, const WebCore::FileSystemHandleIdentifier& targetIdentifier)
        : m_arguments(identifier, targetIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Move {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const WebCore::FileSystemHandleIdentifier&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_Move; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_MoveReply; }
    using AsyncReply = MoveAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const std::optional<WebKit::FileSystemStorageError>& result);
    using Reply = std::tuple<std::optional<WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<std::optional<WebKit::FileSystemStorageError>>;
    Move(const WebCore::FileSystemHandleIdentifier& identifier, const WebCore::FileSystemHandleIdentifier& destinationIdentifier, const String& newName)
        : m_arguments(identifier, destinationIdentifier, newName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetFile {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_GetFile; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<String, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<String, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_GetFileReply; }
    using AsyncReply = GetFileAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const Expected<String, WebKit::FileSystemStorageError>& result);
    using Reply = std::tuple<Expected<String, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<String, WebKit::FileSystemStorageError>>;
    explicit GetFile(const WebCore::FileSystemHandleIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CreateSyncAccessHandle {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_CreateSyncAccessHandle; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<std::pair<WebCore::FileSystemSyncAccessHandleIdentifier, IPC::SharedFileHandle>, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<std::pair<WebCore::FileSystemSyncAccessHandleIdentifier, IPC::SharedFileHandle>, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_CreateSyncAccessHandleReply; }
    using AsyncReply = CreateSyncAccessHandleAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const Expected<std::pair<WebCore::FileSystemSyncAccessHandleIdentifier, IPC::SharedFileHandle>, WebKit::FileSystemStorageError>& result);
    using Reply = std::tuple<Expected<std::pair<WebCore::FileSystemSyncAccessHandleIdentifier, IPC::SharedFileHandle>, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<std::pair<WebCore::FileSystemSyncAccessHandleIdentifier, IPC::SharedFileHandle>, WebKit::FileSystemStorageError>>;
    explicit CreateSyncAccessHandle(const WebCore::FileSystemHandleIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CloseAccessHandle {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const WebCore::FileSystemSyncAccessHandleIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_CloseAccessHandle; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(std::optional<WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_CloseAccessHandleReply; }
    using AsyncReply = CloseAccessHandleAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const std::optional<WebKit::FileSystemStorageError>& result);
    using Reply = std::tuple<std::optional<WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<std::optional<WebKit::FileSystemStorageError>>;
    CloseAccessHandle(const WebCore::FileSystemHandleIdentifier& identifier, const WebCore::FileSystemSyncAccessHandleIdentifier& accessHandleIdentifier)
        : m_arguments(identifier, accessHandleIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetHandleNames {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_GetHandleNames; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<Vector<String>, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<Vector<String>, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_GetHandleNamesReply; }
    using AsyncReply = GetHandleNamesAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const Expected<Vector<String>, WebKit::FileSystemStorageError>& result);
    using Reply = std::tuple<Expected<Vector<String>, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<Vector<String>, WebKit::FileSystemStorageError>>;
    explicit GetHandleNames(const WebCore::FileSystemHandleIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetHandle {
public:
    using Arguments = std::tuple<const WebCore::FileSystemHandleIdentifier&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::NetworkStorageManager_GetHandle; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>&&)>&&);
    static void cancelReply(CompletionHandler<void(Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NetworkStorageManager_GetHandleReply; }
    using AsyncReply = GetHandleAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>& result);
    using Reply = std::tuple<Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>&>;
    using ReplyArguments = std::tuple<Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>>;
    GetHandle(const WebCore::FileSystemHandleIdentifier& identifier, const String& name)
        : m_arguments(identifier, name)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

} // namespace NetworkStorageManager
} // namespace Messages
